package org.bsf.smartValueObject.demo;

import org.bsf.commons.ejb.SessionAdapterBean;
import org.bsf.smartValueObject.mediator.ConcurrencyException;
import org.bsf.smartValueObject.mediator.MediatorException;

import java.util.Calendar;
import java.util.Iterator;
import java.util.TimeZone;

/**
 * This stateful bean is the one used by each client to simulate a
 * persistence service.
 *
 *  @ejb:bean       type="Stateful"
 *                  name="SVODemoUser"
 *                  jndi-name="ejb/SVODemoUser"
 *                  view-type="both"
 *
 *  @ejb:home       local-extends="javax.ejb.EJBLocalHome"
 *                  extends="javax.ejb.EJBHome"
 *
 *  @ejb:interface  local-extends="javax.ejb.EJBLocalObject"
 *                  extends="javax.ejb.EJBObject"
 *
 *  @ejb:transaction type="Required"
 *
 */
public class SVODemoUserBean extends SessionAdapterBean {

    private DemoMediator userMediator;

    /**
     * @ejb:interface-method
     */
    public void ejbCreate() {
        userMediator = new DemoMediator();
        try {
            userMediator.updateGraph( buildUserGraph() );
        } catch( MediatorException e ) {
            handleExceptionAsSystemException( e );
        }
        logInfo( "SVODemoUser created..." );
    }

    /**
     * @ejb:interface-method
     */
    public Object getCompanyVo( Object propotype ) {
        Object result = null;
        try {
            result = userMediator.getGraph( propotype );
        } catch( MediatorException e ) {
            handleExceptionAsSystemException( e );
        }
        return result;
    }

    /**
     * @ejb:interface-method
     */
    public String updateCompanyVo( Object propotype ) throws ConcurrencyException {
        try {
            return userMediator.updateGraphAndReport( propotype );
        } catch( MediatorException e ) {
            if ( e instanceof ConcurrencyException ) {
                ConcurrencyException ce = (ConcurrencyException) e;
                throw ce;
            } else {
                handleExceptionAsSystemException( e );
            }
        }
        return null;
    }

    /**
     * Simulates a concurrent modification on the graph
     *
     * @ejb:interface-method
     */
    public void modifyConcurrently() {
        try {
            CompanyVO vo = new CompanyVO();
            vo.setId( new Long( 0 ) );
            CompanyVO comp = (CompanyVO) userMediator.getGraph( vo );
            Iterator subIt = comp.subsidiaries();
            while ( subIt.hasNext() ) {
                SubsidiaryVO subsidiaryVO = (SubsidiaryVO) subIt.next();
                subsidiaryVO.setName( subsidiaryVO.getName() + " (Modified)" );
                break;
            }
            userMediator.updateGraph( comp );
        } catch( MediatorException e ) {
            handleExceptionAsSystemException( e );
        }

    }

    private static final Object buildUserGraph() {
        CompanyVO comp = new CompanyVO();
        comp.setName( "The big big company" );

        // set time to actual date, 0:00:00
        Calendar cal = Calendar.getInstance( TimeZone.getTimeZone( "Europe/Paris" ) );
        cal.set( Calendar.HOUR_OF_DAY, 0 );
        cal.set( Calendar.MINUTE, 0 );
        cal.set( Calendar.SECOND, 0 );
        cal.set( Calendar.MILLISECOND, 0 );
        comp.setCreationDate( cal.getTime() );

        SubsidiaryVO sub = new SubsidiaryVO();
        sub.setName( "Paris" );
        sub.setWorkforce( 2 );
        comp.addSubsidiary( sub );

        sub = new SubsidiaryVO();
        sub.setName( "London" );
        sub.setWorkforce( 1 );
        comp.addSubsidiary( sub );

        sub = new SubsidiaryVO();
        sub.setName( "Berlin" );
        sub.setWorkforce( 1 );
        comp.addSubsidiary( sub );

        return comp;
    }
}
